"""Constants used by the izone component."""

IZONE = "izone"

DATA_DISCOVERY_SERVICE = "izone_discovery"
DATA_CONFIG = "izone_config"

DISPATCH_CONTROLLER_DISCOVERED = "izone_controller_discovered"
DISPATCH_CONTROLLER_DISCONNECTED = "izone_controller_disconnected"
DISPATCH_CONTROLLER_RECONNECTED = "izone_controller_disconnected"
DISPATCH_CONTROLLER_UPDATE = "izone_controller_update"
DISPATCH_ZONE_UPDATE = "izone_zone_update"

TIMEOUT_DISCOVERY = 20
