"""Define constants for the WWLLN integration."""
from datetime import timedelta

DOMAIN = "wwlln"

CONF_WINDOW = "window"

DATA_CLIENT = "client"

DEFAULT_RADIUS = 25
DEFAULT_WINDOW = timedelta(hours=1)
